/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _MODULEINTERFACE_H
#define _MODULEINTERFACE_H

#include <string>

#include "../EventManager/Event.h"
#include "../Platform.h"

class Entity;

/**
 * Basic interface for modules.
 * Provides methods for initialisation and cleanup.
 */
class INVRS_SYSTEMCORE_API ModuleInterface {
public:
	virtual ~ModuleInterface() {};

	enum CONFIGURATION_ORDER {
		CONFIGURE_BEFORE_CORECOMPONENTS,
		CONFIGURE_AFTER_CORECOMPONENTS
	};

	/**
	 * Instruct the module to read its configuration from a file.
	 * @return true on success.
	 * @param configFile path of the XML-file.
	 */
	virtual bool loadConfig(std::string configFile) = 0;

	/**
	 * Instruct the module to free all memory allocated internally.
	 */
	virtual void cleanup() {};

	/**
	 * Updates the module every frame
	 */
	virtual void update(float dt) {};

	/**
	 * Return the unique name of the module.
	 */
	virtual std::string getName() = 0;

	/**
	 * Define wether the module should be configured before or after the
	 * core components are configured. By default the module is configured
	 * before the core components are configured.
	 */
	virtual CONFIGURATION_ORDER getConfigurationOrder() {
		return CONFIGURE_BEFORE_CORECOMPONENTS;
	} // getConfigurationOrder

	/**
	 * \todo write doku
	 */
	virtual Event* createSyncEvent() {
		return NULL;
	}

};

/* for plugins: */
typedef bool (*registerModule_t)();
typedef const char* (*modulePluginVersion_t)();

/* internal macro to ensure interface-compatibility between SystemCore
 * and plugins. (Is checked on plugin-loading) */
#define MODULEPLUGINVERSION "$Revision: 1622 $"

#ifdef WIN32
#define MAKEMODULEPLUGIN(Module_t, ModuleOwner_t) \
	extern "C" { \
		ModuleInterface* module_ ## Module_t = 0; \
		const char* modulePluginVersion() \
		{ \
			return MODULEPLUGINVERSION; \
		} \
		bool registerModule() \
		{ \
			if (!module_ ## Module_t) { \
				module_ ## Module_t = new Module_t(); /* new could be overloaded outside this compilation unit, so we have to use this function. */ \
				return ModuleOwner_t::registerModuleInterface( module_ ## Module_t ); \
			}\
			return true; /* already registered*/\
		} \
		bool unregisterModule() \
		{ \
			bool ret = false; \
			if (module_ ## Module_t) { \
				printd("%s::unregisterModule(): unregistering Module.\n", #Module_t); \
				ret = ModuleOwner_t::unregisterModuleInterface( module_ ## Module_t ); \
				printd("%s::unregisterModule(): deleting Module.\n", #Module_t); \
				delete module_ ## Module_t; /* delete could be overloaded outside this compilation unit, so we have to use this function. */ \
			} \
			else { \
				printd(ERROR, "%s::unregisterModule(): Module already deleted!\n", #Module_t); \
			} \
			return ret; \
		} \
	}\
	BOOL APIENTRY DllMain( HMODULE hModule, \
						   DWORD  ul_reason_for_call, \
						   LPVOID lpReserved \
						 ) \
	{ \
		switch (ul_reason_for_call)\
		{ \
			case DLL_PROCESS_ATTACH: \
				fprintf(stderr, "%s::DllMain(): process attached.\n", #Module_t); \
				SystemCore::addPreloadModule(hModule);\
				break;\
			case DLL_THREAD_ATTACH: \
				fprintf(stderr, "%s::DllMain(): thread attached.\n", #Module_t); \
				break;\
			case DLL_THREAD_DETACH: \
			case DLL_PROCESS_DETACH: \
				break; \
		} \
		return TRUE; \
	}

#else

#define MAKEMODULEPLUGIN(Module_t, ModuleOwner_t) \
	extern "C" { \
		ModuleInterface* module_ ## Module_t = 0; \
		const char* modulePluginVersion() \
		{ \
			return MODULEPLUGINVERSION; \
		} \
		bool registerModule() \
		{ \
			module_ ## Module_t = new Module_t(); /* new could be overloaded outside this compilation unit, so we have to use this function. */ \
			return ModuleOwner_t::registerModuleInterface( module_ ## Module_t ); \
		} \
		bool unregisterModule() \
		{ \
			bool ret = false; \
			if (module_ ## Module_t) { \
				printd("%s::unregisterModule(): unregistering Module.\n", #Module_t); \
				ret = ModuleOwner_t::unregisterModuleInterface( module_ ## Module_t ); \
				printd("%s::unregisterModule(): deleting Module.\n", #Module_t); \
				delete module_ ## Module_t; /* delete could be overloaded outside this compilation unit, so we have to use this function. */ \
			} \
			else { \
				printd(ERROR, "%s::unregisterModule(): Module already deleted!\n", #Module_t); \
			} \
			return ret; \
		} \
	}
#endif // WIN32

#endif // _MODULEINTERFACE_H
