/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _OLDAPPLICATIONBASE_H
#define _OLDAPPLICATIONBASE_H

#include <gmtl/Quat.h>
#include <gmtl/QuatOps.h>
#include <gmtl/Matrix.h>
#include <gmtl/MatrixOps.h>
#include <gmtl/Generate.h>
#include <gmtl/External/OpenSGConvert.h>

#include <irrXML.h>

#include "SystemCore.h"
#include "WorldDatabase/WorldDatabase.h"
#include "Timer.h"
#include "UserDatabase/User.h"
#include "UserDatabase/UserDatabase.h"
#include "ComponentInterfaces/NetworkInterface.h"
#include "ComponentInterfaces/InteractionInterface.h"
#include "ComponentInterfaces/NavigationInterface.h"
#include "../InputInterface/ControllerManagerInterface.h"
#include "EventManager/EventManager.h"
#include "TransformationManager/TransformationManager.h"
#include "WorldDatabase/CameraTransformation.h"
#include "TransformationManager/TrackingOffsetModifier.h"
#include "TransformationManager/ApplyNavigationModifier.h"
#include "TransformationManager/TransformationDistributionModifier.h"
#include "TransformationManager/UserTransformationWriter.h"
#include "TransformationManager/CameraTransformationWriter.h"
#include "TransformationManager/AvatarTransformationWriter.h"
#include "TransformationManager/EntityTransformationWriter.h"
#include "SystemCore.h"
#include "DebugOutput.h"

/******************************************************************************
 * This class is intended to serve as base class for applications using the
 * inVRs framework. It performs some very basic tasks which should fit to the
 * needs of most applications (like initialising the modules in a specific
 * order, starting the EventManager thread, connecting to an other machine ...).
 * However the OldApplicationBase is still independent from any graphics/window API
 * so any platform specific stuff must be initialised elsewhere. The
 * TutorialBase class is an example for a class which extends the
 * OldApplicationBase by OpenSG specific aspects. If you don't want to use the
 * OldApplicationBase go through some of the tutorials.
 *
 * usage:
 *  1: create an instance of your class (derived from OldApplicationBase)
 *	2: call the <code>run()</code> method
 *
 * <code>run()</code> in turn invokes the following methods in that order:
 *	<code>setupModules()</code>
 *	<code>setupStageModifiers()</code>
 *	<code>init()</code>
 *	<code>setupCamera()</code>  per frame
 *	<code>display()</code>  per frame
 *	<code>cleanup()</code>
 *
 * @deprecated
 */
class INVRS_SYSTEMCORE_API OldApplicationBase {
public:
	/**
	 * The constructor expects in argv[0] a config file which is passed to the
	 * Configuration. This is a good place for performing any platform specific
	 * initialisation. You cannot rely on very much functionality being available
	 * at this stage. It is recommended to do as much as possible of your
	 * initialisation work in <code>init()</code>
	 */
	OldApplicationBase(int argc, char** argv);

	virtual ~OldApplicationBase() {};

	/**
	 * Typically loading the WorldDatabase and similar things are performed here
	 * this method should be overriden by your own method.
	 * @param argc command line argument counter
	 * @param argv command line argument vector
	 * @return returning false leads to program termination
	 */
	virtual bool init(int argc, char** argv);

	/**
	 * This method is invoked periodically. It updates the TransformationManager
	 * and polls the controller. Additionally <code>display()</code> is invoked.
	 * This method is not intended to be overriden. Use <code>display()</code>
	 * instead. The <code>run()</code> method is responsible for calling
	 * <code>step()</code> from time to time.
	 * @param dt elapsed time since last invokation
	 */
	virtual void step(float dt);

	/**
	 * Will be invoked during <code>step()</code>. Some parts of the
	 * transformation pipes might have been already executed.
	 * @param elapsed time since last invokation
	 */
	virtual void display(float dt);

	/**
	 * Thats the proper place to delete stuff allocated by the application.
	 * Modules instanced during setupModules() don't need to be delete, this will
	 * be done automatically.
	 */
	virtual void cleanup();

	/**
	 * All modules and variables which have to be deleted after the WorldDatabase
	 * and UserDatabase are cleaned up can be deleted here.
	 */
	virtual void finalize();

	/**
	 * Override this method to tell the TransformationManager which modifiers to
	 * use. By default the modifiers are loaded from "default.xml" in the
	 * TransformationManager config directory. This method is invoked only once
	 * during startup by <code>run()</code>.
	 *
	 * @deprecated
	 */
	virtual void setupStageModifiers();

	// TODO: fix documentation
	/**
	 * Create instances of the modules you need here. The modules register
	 * themself at the SystemCore automatically. However in order to make the
	 * OldApplicationBase class aware of them its necessary to assign them to the
	 * Interface members for now.
	 * This method will be invoked by <code>run()</code>.
	 * \todo documentation is deprecated -> must be fixed
	 * @param Module module to initialize
	 */
	static void configureModule(ModuleInterface* module);

	// TODO: add documentation
	/**
	 * \todo add documentation
	 */
	virtual void cbConfigureModule(ModuleInterface* module);

	virtual void cbConfigureCoreComponent(CoreComponents component);

	/**
	 * Here the methods responsible for initialisation are invoked in an
	 * appropriate order. Additionally it should make sure <code>step()</code> is
	 * called periodically. The former is already provided by the implementation
	 * of OldApplicationBase.
	 * For instance a glut based derived class could register <code>step()</code>
	 * via <code>glutDisplayFunc()</code> and invoke
	 * <code>glutMainLoop()</code> to make <code>step()</code> called periodically
	 * @param argc command line argument counter
	 * @param argv command line argument vector
	 */
	virtual void run(int argc, char** argv);

	/**
	 * This method must be implemented in a derived class. It should write the
	 * matrix of activeCamera to the 3d engine
	 */
	virtual void setupCamera();

protected:
	friend class SystemCore;

	/**
	 *
	 */
	virtual void updateNavigation(float dt);

	/**
	 *
	 */
	virtual gmtl::Matrix44f& getCameraMatrix(gmtl::Matrix44f& dst);

	/**
	 *
	 */
	static void configureCoreComponent(CoreComponents component);

	/**
	 *
	 */
	virtual void internalCleanup();

	static OldApplicationBase* currentInstance;

	ControllerManagerInterface* contInt;
	NavigationInterface* navInt;
	NetworkInterface* netInt;
	InteractionInterface* interactionInt;

	User* localUser;
	CameraTransformation* activeCamera;
	FILE* logFile;
	float physicalToWorldScale;

};

#endif
