/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _UTILITYFUNCTIONS_H_
#define _UTILITYFUNCTIONS_H_

#include <sstream>

#include "Configuration.h"
#include "ModuleIds.h"
#include "TransformationManager/TransformationManager.h"

/**
 * Returns the full path to the config file. The following concepts are used for
 * creation of the final path:
 *
 * <table>
 * <tr>
 *  <th> path                 </th><th> configFileEntry         </th><th> RESULT </th>
 * </tr> <tr>
 *  <td> "abc.xml"            </td><td> "..."                   </td><td> configFilePath + path </td>
 * </tr> <tr>
 *  <td> "path/to/abc.xml"    </td><td> "..."                   </td><td> configFilePath + path </td>
 * </tr> <tr>
 *  <td> "abc.xml"            </td><td> "" || <code>NULL</code> </td><td> configRootPath + path </td>
 * </tr> <tr>
 *  <td> "path/to/abc.xml"    </td><td> "" || <code>NULL</code> </td><td> configRootPath + path </td>
 * </tr> <tr>
 *  <td> "/path/to/abc.xml"   </td><td> "" || <code>NULL</code> </td><td> path (global path)    </td>
 * </tr> <tr>
 *  <td> "./abc.xml"          </td><td> "" || <code>NULL</code> </td><td> PWD + path            </td>
 * </tr> <tr>
 *  <td> "./path/to/abc.xml"  </td><td> "" || <code>NULL</code> </td><td> PWD + path            </td>
 * </tr> <tr>
 *  <td> "../path/to/abc.xml" </td><td> "" || <code>NULL</code> </td><td> PWD + path            </td>
 * </tr>
 * </table>
 *
 */
inline std::string getFullPath(const std::string& path, std::string configFileEntry = "") {
	std::string result("");

	if (path.length() == 0)
		return result;

#ifdef WIN32
	if (path[0] != '.' && path[0] != '\\' && path[1] != ':') {
#else
	if (path[0] != '.' && path[0] != '/') {
#endif
		if (configFileEntry.length() > 0)
			result = Configuration::getPath(configFileEntry);
		else
			result = Configuration::getPath("root");
	} // if
	result.append(path);

	return result;
} // getFullPath

inline std::string getTransformationPipeIdAsString(OSG::UInt64 id) {
	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	unsigned srcId, dstId, pipeType, objectClass, objectType, objectId;
	bool fromNetwork;

	TransformationManager::unpackPipeId(id, &srcId, &dstId, &pipeType, &objectClass, &objectType, &objectId, &fromNetwork);

	ss << "Source Module: " << getModuleNameById(srcId) << std::endl;
	ss << "Destination Module: " << getModuleNameById(dstId) << std::endl;
	ss << "Pipe Type: " << pipeType << std::endl;
	ss << "Object Class: " << objectClass << std::endl;
	ss << "Object Type: " << objectType << std::endl;
	ss << "Object Id: " << objectId << std::endl;
	ss << "From Network: " << (fromNetwork ? "true" : "false") << std::endl;

	return ss.str();
} // getTransformationPipeIdAsString

inline bool copyFile(const std::string& srcFile, const std::string dstFile) {

	long size;
	char *buffer;
	size_t result;
	bool success = true;
	FILE* in = fopen(srcFile.c_str(), "r");
	FILE* out = fopen(dstFile.c_str(), "w");

	// check if files handles could be opened
	if (!in || !out) {
		if (in)
			fclose(in);
		if (out)
			fclose(out);
		return false;
	} // if

	// obtain file size:
	fseek(in, 0, SEEK_END);
	size = ftell(in);
	rewind(in);

	// allocate memory to contain the whole file:
	buffer = (char*)malloc(sizeof(char) * size);
	if (buffer == NULL) {
		fclose(in);
		fclose(out);
		return false;
	} // if

	// copy the file into the buffer:
	result = fread(buffer, 1, size, in);
	if (result != size) {
		fclose(in);
		fclose(out);
		free(buffer);
		return false;
	} // if

	// write file into file
	result = fwrite(buffer, 1, size, out);
	if (result != size) {
		success = false;
	} // if

	// terminate
	fclose(in);
	fclose(out);
	free(buffer);

	return success;
} // copyFile

#endif /*_UTILITYFUNCTIONS_H_*/
