/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
 \*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _AVATARAAVATAR_H
#define _AVATARAAVATAR_H

#include <string>

#include <OpenSG/OSGNode.h>
#include <OpenSG/OSGTransform.h>

#include <avatara/OSGAvatar.h>

#include <inVRs/SystemCore/DataTypes.h>
#include <inVRs/SystemCore/ComponentInterfaces/AvatarInterface.h>
#include <inVRs/OutputInterface/OpenSGSceneGraphInterface/OpenSGSceneGraphInterface.h>

#ifdef WIN32
	#ifdef INVRSAVATARAWRAPPER_EXPORTS
	#define INVRS_AVATARAWRAPPER_API __declspec(dllexport)
	#else
	#define INVRS_AVATARAWRAPPER_API __declspec(dllimport)
	#endif
#else
	#define INVRS_AVATARAWRAPPER_API
#endif


/******************************************************************************
 * Wrapper class to the avatara library to provide avatara avatars in inVRs.
 * This class allows to integrate an avatar from the avatara library into
 * inVRs.
 */
class INVRS_AVATARAWRAPPER_API AvataraAvatar : public AvatarInterface {
public:

	/**
	 * Constructor
	 */
	AvataraAvatar();

	/**
	 * Destructor
	 */
	~AvataraAvatar();

	/**
	 * Loads the avatar from the passed xml stream
	 */
	virtual bool loadConfig(std::string configFile);

	/**
	 * Returns the model of the Avatar
	 */
	virtual ModelInterface* getModel();

	/**
	 * Sets the visibility of the avatar
	 */
	virtual void showAvatar(bool active);

	/**
	 * Returns if the avatar is visible or not
	 */
	virtual bool isVisible();

	/**
	 * Sets the transformation of the avatar
	 */
	virtual void setTransformation(TransformationData trans);

	/**
	 * Updates the avatar model by calculating the bone transformations and
	 * correcting the model mesh.
	 */
	virtual void update(float dt);

	/**
	 * Starts an animation loop using the animation with the passed name.
	 */
	bool startAnimation(std::string animationType, int transitionTime = 0);

	/**
	 * Starts the animation once with the passed name.
	 */
	bool runAnimationOnce(std::string animationType, int transitionTime = 0);

	/**
	 * Stops the current animation.
	 */
	void stopAnimation();

private:
	/**
	 * Sets the user the avatar corresponds to
	 */
	virtual void setOwner(User* user);

	/**
	 * Loads the configuration from the passed config file
	 */
//	bool loadConfig(std::string configFile);

	/**
	 * Creates the OpenSG-representation of the avatar
	 */
	OSG::NodePtr buildAvatar();

	/**
	 * Sets the avatar's hand transformation
	 */
	void setHandTransformation(TransformationData trans);

	/**
	 * Sets the avatar's head transformation
	 */
	void setHeadTransformation(TransformationData trans);

	/**
	 * Sets the avatar's view direction
	 */
	void setViewDirection(gmtl::Vec3f dir);

	static const int defaultTransitionTime;	// defines the default transition time to reach idle

	bool smoothAnimation;			// defines if the animations should be smooth
	float speed;					// sets the speed of the animations
	bool singleRunAnimation;		// defines if a single animation is running
	int singleAnimationTime;		// current time of the animation
	std::string imageFile;			// url to the avatar texture
	std::string defaultAnimation;	// name of the default avatar animation
	float avatarScale;				// scale of the avatar
	bool avatarInitialized;			// stores if the avatar is initialized
	bool avatarGeometryBuilt;		// stores if the avatar's geometry is built

	TransformationData modelTransform;	// model transformation (from config file)

	/// OpenSG Avatar.
	Avatara::OSGAvatar avatar;
	/// Avatar model.
	Avatara::Model model;
	/// Avatar animation.
	std::vector<Avatara::Animation*> animations;
	/// Animation set for all avatar animations.
	Avatara::AnimationSet animationSet;

	OSG::TransformPtr avatarOffsetCore;	// OpenSG core for avatar offset node
	OSG::TransformPtr avatarTrans;		// OpenSG core for avatar transformation node
	OSG::NodePtr avatarNode;			// OpenSG node for avatar model
	OpenSGModel* avatarModel;			// OpenSGModel for avatar

//*****************************************************************************
// Configuration loading
//*****************************************************************************
private:
	static XmlConfigurationLoader xmlConfigLoader;

	class ConverterToV1_0a4 : public XmlConfigurationConverter {
	public:
		ConverterToV1_0a4();
		virtual bool convert(XmlDocument* document, const Version& version, std::string configFile);
	}; // ConverterToV1_0a4
}; // AvataraAvatar

/******************************************************************************
 * Factory class for the AvataraAvatar
 */
class INVRS_AVATARAWRAPPER_API AvataraAvatarFactory : public AvatarFactory {
public:
	/**
	 * Creates a new AvataraAvatar if the className matches
	 */
	AvatarInterface* create(std::string className, std::string configFile);
}; // AvataraAvatarFactory

#endif //_AVATARAAVATAR_H
