/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                        Project: CAVESceneManager                          *
 *                                                                           *
 * The CAVE Scene Manager (CSM) is a wrapper for the OpenSG multi-display    *
 * functionality, which allows the easy configuration of a multi-display     *
 * setup via text files. It was developed by Adrian Haffegee of the Centre   *
 * for Advanced Computing and Emerging Technologies (ACET) at the            *
 * University of Reading, UK as part of his MSc Thesis.                      *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*! \file OSGCAVEConfig.h
 *
 *	Adrian Haffegee, Sept 2004
 *	Header file for CAVE Scene Manager (CSM) configurator class
 */

#ifndef _CVR_CAVECONFIG_H_
#define _CVR_CAVECONFIG_H_


#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGGeometry.h>
#include "OSGCAVEWall.h"

OSG_BEGIN_NAMESPACE

//@RLANDER: added definitions for windows-dll compatibility
#if defined(WIN32)
	#ifdef CAVESCENEMANAGER_COMPILELIB
		#define CAVESCENEMANAGER_DLLMAPPING     __declspec(dllexport)
	#else
		#define CAVESCENEMANAGER_DLLMAPPING     __declspec(dllimport)
	#endif
#else
	#define CAVESCENEMANAGER_DLLMAPPING
#endif

#define MAXCAVEWALLS 10

/*! \brief Different types measurements
 */
enum CAVEUnitType{  CAVEUnitFeet   =  0, /**< Feet */
	CAVEUnitMeters,			/**< Meters */
	CAVEUnitInches,			/**< Inches */
	CAVEUnitCentimeters,	/**< Centimeters */
	CAVEUnit_Last
};


/*! \brief Names and conversion factors for measurements
 */
struct CAVEUnitsStruct{
	CAVEUnitType type;		/**< type */
	char* name;				/**< name */
	float cF;				/**< conversion factor (to cm) */
};

/*! \brief Struct for configuration functions based on names
 */
struct cfgFuncStruct{
	char* cfgParam;									/**< parameter name */
	bool (*cfgFunc)(struct CAVEConfig *, char*);	/**< config function */
};

/*! \brief Configurator class for CSM
 */
class CAVESCENEMANAGER_DLLMAPPING CAVEConfig {
	public:

		CAVEConfig();
		virtual ~CAVEConfig(void);

		bool loadFile(char* cfgFile);
		void printConfig();
		bool checkConfig();

		void setStereo(bool useStereo);
		void setMCastServer(bool useMCast);
		void setEyeSeparation(float eyeSep, CAVEUnitType units);
		void setTrackerKey(int key);
		void setControllerKey(int key);
		void setCAVEHeight(float height, CAVEUnitType units);
		void setCAVEWidth(float width, CAVEUnitType units);
		void setCAVEOrigin(Pnt3f origin, CAVEUnitType units);

		bool isStereo();
		bool isMCastServer();
		float getEyeSeparation();
		int getTrackerKey();
		int getControllerKey();
		int getNumActiveWalls();
		CAVEWall* getActiveWall(int wallNum);
		int getNumPipes();

	private:

		bool parseLine(char* line);
		static bool parseDisplayMode(CAVEConfig *cfgObj, char* params);
		static bool parseEyeSparation(CAVEConfig *cfgObj, char* params);
		static bool parseMulticastServerComms(CAVEConfig *cfgObj, char* params);
		static bool parseTrackerDaemonKey(CAVEConfig *cfgObj, char* params);
		static bool parseControllerDaemonKey(CAVEConfig *cfgObj, char* params);
		static bool parseWalls(CAVEConfig *cfgObj, char* params);
		static bool parseWallDisplay(CAVEConfig *cfgObj, char* params);
		static bool parseProjectionData(CAVEConfig *cfgObj, char* params);
		static bool parseOrigin(CAVEConfig *cfgObj, char* params);
		static bool parseCAVEWidth(CAVEConfig *cfgObj, char* params);
		static bool parseCAVEHeight(CAVEConfig *cfgObj, char* params);

		static char* skipWhiteSpace(char* data);
		static char* getNextFloat(char *data, float* value);
		static char* getNextInt(char *data, int* value);
		static char* getNextString(char *data, char* string, int length);


		bool m_stereo;
		bool m_mCastServer;
		float m_eyeSeparation;
		int m_trackerKey;
		int m_ControllerKey;
		CAVEWall* m_walls[MAXCAVEWALLS];
		int m_activeWalls;
		CAVEUnitType m_origUnits;
		Pnt3f m_origin;
		CAVEUnitType m_heightUnits;
		float m_height;
		CAVEUnitType m_widthUnits;
		float m_width;
		int m_numPipes;

		static struct cfgFuncStruct cfgFunctions[];
		static struct CAVEUnitsStruct CAVEUnits[CAVEUnit_Last];

};

OSG_END_NAMESPACE


#endif // _CVR_CAVECONFIG_H_
